# Copyright 2019, 2020 razer <razerraz@free.fr>
# Copyright 2016 Timo Kankare <timo.kankare@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.

"""Backends to implement mail box specific functionality, like IMAP and POP3."""

from abc import ABCMeta, abstractmethod
import re
import email
import logging
import imaplib
import poplib
from ssl import SSLError
import errno
import socket

from .account import RemoteAccount as RA
from .utils import utf7_decode, utf7_encode
from .i18n import _

IMAP = 'imap'
POP3 = 'pop3'
BACKEND_MAP = {IMAP: imaplib.IMAP4, POP3: poplib.POP3}
SSL_BACKEND_MAP = {IMAP: imaplib.IMAP4_SSL, POP3: poplib.POP3_SSL}


class RemoteMail:
    """Interface for mailbox backends.
    Mailbox backend implements access to the specific type of mailbox.
    """

    __metaclass__ = ABCMeta

    UNKNOWN = 0
    FAILED = 1
    REFUSED = 2
    BAD_LOGIN = 3
    UNSECURE = 4
    CONNECT_ERROR = 5
    TIMEOUT = 6
    SUCCESS = 7
    OFFLINE = -1
    NOCRED_ERR = _('No credentials found !')
    AUTH_ERR = _('Authentification failed')
    INSECURE_ERR = _('Insecure connection')
    TIMEOUT_ERR = _('Connection timeout')
    REFUSED_ERR = _('Connection refused')
    UNKNOWN_ERR = _('Unknown error')

    def __init__(self, account):
        self.account = account
        self.backend = BACKEND_MAP[self.account[RA.BACKEND]]
        self.ssl_backend = SSL_BACKEND_MAP[self.account[RA.BACKEND]]
        self.connection = None
        self.err_str = f'{self.account[RA.NAME]}({self.account[RA.SERVER]}):'
        socket.setdefaulttimeout(10)

    def _safe_call(self, command, *args, **kwargs):
        """ Try / except wrapper for remote commands and its arguments
            Filter out 'set_status' and 'exc_info' from keyword args
            for internal purpose
            Set account status on error if 'set_status' is True
            Log execution info on error if 'exc_info' is True
        """
        set_status = True
        exc_info = True
        if 'set_status' in kwargs.keys():
            set_status = kwargs['set_status']
            kwargs.pop('set_status')
        if 'exc_info' in kwargs.keys():
            exc_info = kwargs['exc_info']
            kwargs.pop('exc_info')

        def set_account_status(status, error):
            if set_status:
                self.account.set_status(status, error)

        try:
            return self.SUCCESS, command(*args, **kwargs)
        except socket.timeout:
            logging.error(f'{self.err_str} timeout')
            set_account_status(self.TIMEOUT, self.TIMEOUT_ERR)
            return self.TIMEOUT, None
        except imaplib.IMAP4.error:
            set_account_status(self.CONNECT_ERROR, self.UNKNOWN_ERR)
            return self.CONNECT_ERROR, None
        except Exception:
            logging.error(f'{self.err_str} unknown error', exc_info=exc_info)
            set_account_status(self.CONNECT_ERROR, self.UNKNOWN_ERR)
            return self.UNKNOWN, None

    def open(self):  # pylint: disable=R0912, R0915
        """ Opens the mailbox
        """
        if self.account.is_locked:
            logging.warning(f'Account {self.account[RA.NAME]}: avoiding '
                            + 'connect attempt without credentials !!!')
            self.account.set_status(self.FAILED, self.NOCRED_ERR)
            return None
        account_name = self.account[RA.NAME]
        server = self.account[RA.SERVER]
        port = self.account[RA.PORT]
        if self.connection:
            logging.warning(f'{account_name}({self.account[RA.TYPE]}): '
                            + f'connection on {server} is already open')
            return self.connection

        def _open(backend, server, port):
            if port:
                self.connection = backend(server, port)
            else:
                self.connection = backend(server)

        try:
            _open(self.ssl_backend, server, port)
        except SSLError:
            try:
                _open(self.backend, server, port)
            except ConnectionRefusedError:
                self.account.set_status(self.BAD_LOGIN, self.REFUSED_ERR)
            except socket.gaierror:
                self.account.set_status(self.REFUSED, self.REFUSED_ERR)
            else:
                self.account.set_status(self.UNSECURE, self.INSECURE_ERR)
        except ConnectionRefusedError:
            self.account.set_status(self.BAD_LOGIN, self.REFUSED_ERR)
        except (socket.gaierror, OSError) as oserror:
            status = self.OFFLINE
            error = None
            if oserror.errno not in (errno.ENETDOWN, errno.ENETUNREACH):
                status = self.CONNECT_ERROR
                error = oserror.strerror
            self.account.set_status(status, error)
        except Exception:
            logging.error('Unhandled exception', exc_info=True)
            self.account.set_status(self.UNKNOWN_ERR, 'Unhandled exception')
        else:
            return self.connection
        status = self.account[RA.STATUS]
        log_func = logging.error
        if status in (self.UNSECURE, self.OFFLINE):
            log_func = logging.warning
        log_func(f'{self.account[RA.BACKEND]} server {server} from '
                 + f'account {account_name}: {self.account[RA.ERROR]}')
        return self.connection

    def decode_message(self, message):
        try:
            return email.message_from_bytes(message)
        except Exception:
            logging.error(
                f'Decoding email from {self.account[RA.NAME]}: {message}',
                exc_info=True)
        return None

    @abstractmethod
    def close(self):
        """ Closes the mailbox.
        """
        if not self.connection:
            return None
        return True

    @abstractmethod
    def is_open(self):
        """ Returns true if mailbox is open.
        """
        raise NotImplementedError

    @abstractmethod
    def list_messages(self):
        """ Lists unseen messages from the mailbox for this account.
            Yields tuples (folder, message) for every message.
            Return None on error
            Be aware that return value need to be iterable
        """
        self.connection = self.connection if self.connection else self.open()
        if not self.is_open():
            return None
        return True

    @abstractmethod
    def request_folders(self):
        """ Returns list of folder names available in the mailbox.
            Raises an exception if mailbox does not support folders.
        """
        raise NotImplementedError


class RemoteImap(RemoteMail):
    """ Implementation of IMAP mail boxes.
    """
    MISSING_FOLDER = -1
    def __init__(self, account):
        super().__init__(account)
        self.separator = '.'

    def open(self):
        if not super().open():
            return None
        if RA.OAUTH2 in self.account and self.account[RA.OAUTH2]:
            status, _unused = self._safe_call(self.connection.authenticate,
                                              'XOAUTH2',
                                              lambda x: self.account[RA.OAUTH2],
                                              set_status=False)
        else:
            status, _unused = self._safe_call(self.connection.login,
                                              self.account[RA.USER],
                                              self.account[RA.PASS],
                                              set_status=False)
        if status != self.SUCCESS:
            if status == self.CONNECT_ERROR:
                self.account.set_status(self.BAD_LOGIN, self.AUTH_ERR)
            elif status == self.TIMEOUT:
                self.account.set_status(self.TIMEOUT, self.TIMEOUT_ERR)
            else:
                self.account.set_status(status, self.UNKNOWN_ERR)
            return None
        self.connection.utf8_enabled = True
        self.account.set_status(self.SUCCESS, '')
        status, response = self._safe_call(self.connection.namespace,
                                           set_status=False, exc_info=False)
        if status == self.SUCCESS and response and response[0] == 'OK':
            try:
                self.separator = str(response[1]).split()[1][1]
            except IndexError:
                pass
        return self.connection

    def close(self):
        if not super().close():
            return
        for method in 'select', 'close', 'logout':
            status, _unused = self._safe_call(getattr(self.connection, method),
                                              set_status=False, exc_info=False)
            if status != self.SUCCESS:
                break
        self.connection = None

    def is_open(self):
        return self.connection and self.connection.state != 'LOGOUT'

    def select_folder(self, folder):
        status, result = self._safe_call(
            self.connection.select, f'"{utf7_encode(folder)}"', readonly=True)
        if status == self.TIMEOUT:
            return self.TIMEOUT
        if status != self.SUCCESS or not result:
            return self.UNKNOWN_ERR
        if result[0] != 'OK':
            return self.MISSING_FOLDER
        return self.SUCCESS

    def fetch(self, folder, num):
        result = list()
        status, (_unused, msg_data) = self._safe_call(
            self.connection.fetch, num, '(BODY.PEEK[HEADER])')
        if status != self.SUCCESS:
            return result
        for part in msg_data:
            if (not isinstance(part, (tuple, list)) or len(part) < 2
                    or not isinstance(part[1], (bytes, bytearray))):
                continue
            message = self.decode_message(part[1])
            if message:
                result.append((folder, message))
        return result

    def list_messages(self):
        if not super().list_messages():
            return None
        mail_map = []
        folder_map = ['INBOX']
        missing_folders = list()
        if RA.FOLDERS in self.account and self.account[RA.FOLDERS]:
            folder_map = self.account[RA.FOLDERS]
        for folder in folder_map:
            folder_select = self.select_folder(folder)
            if folder_select == self.TIMEOUT:
                break
            if folder_select == self.MISSING_FOLDER:
                missing_folders.append(folder)
            if folder_select != self.SUCCESS:
                continue
            status, response = self._safe_call(self.connection.search,
                                               None, 'UNSEEN')
            if status == self.TIMEOUT:
                return None
            if status != self.SUCCESS:
                continue
            if not response or len(response) < 2:
                logging.error(
                    f'Malformed response from {self.account[RA.SERVER]}')
                continue
            result, data = response
            if result != 'OK' or data == b'':
                logging.debug(f'Folder:{folder}, status:{status}, data:{data}')
                continue # Bugfix LP-735071
            for num in data[0].split():
                mail_map = mail_map + self.fetch(folder, num)
        if missing_folders:
            logging.warning(f'Folder(s) {missing_folders} missing, removing')
            for folder in missing_folders:
                self.account[RA.FOLDERS].remove(folder)
            self.account.save(create=False)
        return mail_map

    def request_folders(self):
        self.connection = self.connection if self.connection else self.open()
        if not self.is_open():
            return []
        status, (_unused, data_map) = self._safe_call(self.connection.list)
        if status != self.SUCCESS or not data_map:
            return []
        self.close()
        folder_list = list()
        for data in data_map:
            match = re.match(r'.+\s+("."|"?NIL"?)\s+"?([^"]+)"?$',
                             data.decode('utf-8'))
            if not match:
                logging.warning(f'{self.account[RA.SERVER]}: bad folder format')
                continue
            folder = utf7_decode(match.group(2))
            folder_list.append(folder)
        return folder_list


class RemotePop3(RemoteMail):
    """ Implementation of POP3 mail boxes.
    """
    def open(self):
        if not super().open():
            return None
        try:
            self.connection.getwelcome()
            self.connection.user(self.account[RA.USER])
            self.connection.pass_(self.account[RA.PASS])
        except poplib.error_proto:
            logging.error(f'{self.err_str} bad credentials')
            self.account.set_status(self.BAD_LOGIN, self.AUTH_ERR)
        except socket.timeout:
            logging.error(f'{self.err_str} timeout')
            self.account.set_status(self.TIMEOUT, self.TIMEOUT_ERR)
        except Exception:
            logging.error(f'{self.err_str} unhandled error', exc_info=True)
            self.account.set_status(self.CONNECT_ERROR, self.UNKNOWN_ERR)
        else:
            if self.account[RA.STATUS] != self.UNSECURE:
                self.account.set_status(self.SUCCESS, '')
            return self.connection
        return None

    def close(self):
        if not super().close():
            return
        self._safe_call(self.connection.quit, set_status=False, exc_info=False)
        self.connection = None

    def is_open(self):
        return self.connection and 'sock' in self.connection.__dict__

    def list_messages(self):
        if not super().list_messages():
            return None
        folder = ''
        mail_map = list()
        status, result = self._safe_call(self.connection.list)
        if status != self.SUCCESS or not len(result) > 1 or not result[1]:
            return None
        mail_total = len(result[1])
        for i in range(1, mail_total + 1):
            status, result = self._safe_call(self.connection.top, i, 0)
            if status == self.TIMEOUT:
                return None
            if status != self.SUCCESS or not len(result) > 1 or not result[1]:
                continue
            message = self.decode_message(b'\n'.join(result[1]))
            if message:
                mail_map.append((folder, message))
        return mail_map

    def request_folders(self):
        raise NotImplementedError('POP3 does not support folders')
